﻿using System;
using System.Collections.ObjectModel;
using System.IO;
using System.Windows.Forms;
using ThinkGeo.Core;

namespace CreateLayerFromIntersect
{
    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();
        }

        /// <summary>
        /// This demo shows how to find intersecting features across two different layers with one layer used as the
        /// target layer. The results of the intersect is then used to create a new layer.
        ///
        /// In a thematic sense, the demo creates a new layer of all countries affected by disputed areas.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Form1_Load(object sender, EventArgs e)
        {
            // Load countries layer, ensure index exists, and style layer
            var countriesPath = @"data\ne_50m_admin_0_countries.shp";
            ShapeFileFeatureSource.BuildIndexFile(countriesPath);
            var countriesLayer = new ShapeFileFeatureLayer(countriesPath);
            countriesLayer.ZoomLevelSet.ZoomLevel01.DefaultAreaStyle = AreaStyle.CreateSimpleAreaStyle(GeoColors.WhiteSmoke, GeoColors.LightGray);
            countriesLayer.ZoomLevelSet.ZoomLevel01.ApplyUntilZoomLevel = ApplyUntilZoomLevel.Level20;

            // Load disputedAreas layer, ensure index exists, and style layer
            var disputedAreasPath = @"data\ne_50m_admin_0_breakaway_disputed_areas.shp";
            ShapeFileFeatureSource.BuildIndexFile(disputedAreasPath);
            var disputedAreasLayer = new ShapeFileFeatureLayer(disputedAreasPath);
            disputedAreasLayer.ZoomLevelSet.ZoomLevel01.DefaultAreaStyle = AreaStyle.CreateSimpleAreaStyle(GeoColors.Red);
            disputedAreasLayer.ZoomLevelSet.ZoomLevel01.ApplyUntilZoomLevel = ApplyUntilZoomLevel.Level20;

            // Get the intersecting features between the first two layers
            var targetShape = GetTargetShape(disputedAreasLayer);
            var intersectingFeatures = GetIntersectingFeatures(countriesLayer, targetShape);
            countriesLayer.Open();
            var columns = ((ShapeFileFeatureSource) countriesLayer.FeatureSource).GetDbfColumns();
            countriesLayer.Close();

            // Create a new layer from the intersection query and style layer
            var intersectionLayer = CreateIntersectionLayer(intersectingFeatures, columns);
            intersectionLayer.ZoomLevelSet.ZoomLevel01.DefaultAreaStyle = AreaStyle.CreateSimpleAreaStyle(GeoColors.Green);
            intersectionLayer.ZoomLevelSet.ZoomLevel01.ApplyUntilZoomLevel = ApplyUntilZoomLevel.Level20;

            // Add the layers to the overlay and add overlay to map
            var overlay = new LayerOverlay();
            overlay.Layers.Add(countriesLayer);
            overlay.Layers.Add(intersectionLayer);
            overlay.Layers.Add(disputedAreasLayer);
            mapView.Overlays.Add(overlay);
        }

        /// <summary>
        /// Generates a multipolygon shape that is a union of all features from the target layer
        ///
        /// WARNING: large data sources may affect performance
        /// </summary>
        /// <param name="targetLayer">The layer whose features will be unioned together to create a single shape</param>
        /// <returns>The union of all features from the layer</returns>
        private MultipolygonShape GetTargetShape(FeatureLayer targetLayer)
        {
            targetLayer.Open();
            var allFeatures = targetLayer.FeatureSource.GetAllFeatures(ReturningColumnsType.NoColumns);
            var targetShape = AreaBaseShape.Union(allFeatures);
            targetLayer.Close();

            return targetShape;
        }

        /// <summary>
        /// Gets the features that intersect with the target area
        /// </summary>
        /// <param name="layer">The layer whose features we want to query for</param>
        /// <param name="area">The area used for the intersect</param>
        /// <returns>A collection of features that intersect the area</returns>
        private Collection<Feature> GetIntersectingFeatures(FeatureLayer layer, AreaBaseShape area)
        {
            layer.Open();
            var features = layer.QueryTools.GetFeaturesIntersecting(area, ReturningColumnsType.AllColumns);
            layer.Close();
            return features;
        }

        /// <summary>
        /// Create a new shapefile that contains all the features from the intersect query. It's file location will be
        /// in the bin folder.
        /// </summary>
        /// <param name="intersectingFeatures">Collection of features found from the intersect query</param>
        /// <param name="columns">The database column names</param>
        /// <returns>The created layer</returns>
        private FeatureLayer CreateIntersectionLayer(Collection<Feature> intersectingFeatures, Collection<DbfColumn> columns)
        {
            // Create a new layer
            var intersectingFeaturesPath = @"data/intersectingFeatures.shp";
            ShapeFileFeatureSource.CreateShapeFile(ShapeFileType.Polygon, intersectingFeaturesPath, columns);
            var intersectionLayer = new ShapeFileFeatureLayer(intersectingFeaturesPath, FileAccess.ReadWrite);

            // Add new features to the layer
            intersectionLayer.Open();
            intersectionLayer.FeatureSource.BeginTransaction();
            foreach (var feature in intersectingFeatures)
            {
                intersectionLayer.FeatureSource.AddFeature(feature);
            }
            intersectionLayer.FeatureSource.CommitTransaction();
            intersectionLayer.Close();

            // Build the shapefile index
            ShapeFileFeatureSource.BuildIndexFile(intersectingFeaturesPath);

            return intersectionLayer;
        }
    }
}
